%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Generation of analysis and synthesis atoms of an ERB-MDCT    %
% and illustration of several properties                       %
% Olivier Derrien - LMA/CNRS - 02.2015                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear all;
close all;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Basic constants
Fs = 44100; % sampling frequency (in Hz)
v = 1; % number of bands per ERB (positive integer)
L = 4092; % length of the time-domain (in samples)
% but the effecive value is usually slightly different
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Generation of the ERB scale
dE = 1/v;
Imax = round(v*9.265*log(1+Fs/2/228.8455));
Emax = Imax/v;
alpha = Fs/2/(exp(Emax/9.265)-1);
E = dE:dE:Emax;
f_center = alpha*(exp(E/9.265)-1);
k_center = 2*L*f_center/Fs;

% Computing the bandwidths (in MDCT index)
k = [k_center(1), k_center];
K = length(k);
Delta = ones(K,K);
Delta = tril(Delta);
Delta(2:end,1) = 0.5;
temp = diag(0.5*ones(1,K));
Delta = Delta-temp;
Delta(1,1) = 1;
Nr = inv(Delta)*k';
N = round(Nr/2)*2;
Nt = Delta(end,:)*N; % effective length of the atoms
N = N';

% Check if this solution is valid
if (sum(diff(N)<0) | sum(N==0))
   fprintf('No valid solution for this configuration\n')
   fprintf('Try with a longer time-domain of less bands per ERB\n')
   break
end

fprintf('The effecive length of atoms is %d\n',Nt)
fprintf('Total number of transform coefficients is %d\n',sum(N))
fprintf('Redundancy is %e\n',sum(N)/Nt)

% Plot the number of transform coefficients in each band
figure(1)
stem(N)
xlabel('Frequency band index (p)')
ylabel('Transform coeffcients per band (N_p)')

% Compute intermediate constants for variable-size MDCT
Ng = [N(1) N(1:end-1)]; % length of sine archs, left part
Nd = N; % length of sine archs, right part
N1 = [0 (N(2:end)-N(1:end-1))/2]; % number of ones in the middle
Nf = length(N); % total number of frequency bands
ofs=zeros(1,Nf); % offset of the windows in the frequency domain
for p = 2:Nf
   ofs(p) = ofs(p-1)+N(p-1)-N1(p); 
end
Nt_pad = ofs(Nf)+2*N1(Nf)+N(Nf-1)+N(Nf); % total length of atoms with zero padding
zg = ofs+N1; % length of zero padding, left part
zd = Nt_pad-(zg+Ng+Nd+N1); % length of zero padding, right part

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Generation of windows in the frequency domain
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

w = zeros(Nf,Nt_pad);
for p = 1 : Nf
    w(p,:) = [zeros(1,zg(p)), sin(pi/2/Ng(p)*([0:Ng(p)-1]+0.5)), ones(1,N1(p)), cos(pi/2/Nd(p)*([0:Nd(p)-1]+0.5)), zeros(1,zd(p))];
end

% Plot the windows
figure(2)
plot(w')
xlabel('Frequency index (p)')
ylabel('MDCT windows (w_p)')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Computing the variable-length MDCT atoms in the frequency domain
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

k = [0:Nt_pad-1];
D_ERB = zeros(Nt_pad,sum(N));
idx = 1;
for p = 1 : Nf
    for n = 0 : N(p)-1
        phi = sqrt(2/N(p)) * w(p,:) .* cos(pi/N(p)*(k-ofs(p)+N(p)/2+0.5)*(n+0.5));
        D_ERB(:,idx) = phi';
        idx = idx+1;
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Specification of the DCT-IV as a matrix
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

t_DCT = [0:Nt-1];
f_DCT = [-N(1):Nt+N(end)-1];
D_DCT = sqrt(2/Nt) * cos(pi/Nt*(t_DCT'+0.5)*(f_DCT+0.5));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Computing ERB-MDCT analysis and synthesis dictionaries
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

DictA = D_DCT * D_ERB;
DictS = D_ERB(N(1)+1:end-N(end),:)' * D_DCT(:,N(1)+1:end-N(end))';

% Plot magnitude of analysis dictionary as an image
figure(3)
imagesc(abs(DictA'));
colorbar
xlabel('Samples')
ylabel('Atom index')
title('Magnitude of analysis dictionary')

% Plot magnitude of synthesis dictionary as an image
figure(4)
imagesc(abs(DictS));
colorbar
xlabel('Samples')
ylabel('Atom index')
title('Magnitude of synthesis dictionary')

% Checking perfect reconstruction capability
I = DictA * DictS;

figure(5)
imagesc(I)
colorbar
xlabel('Input samples')
ylabel('Output samples')
title('Analysis-synthesis operator matrix')

% Computing the scalar products for analysis and synthesis atoms
SpA = DictA'*DictA;
SpS = DictS*DictS';

% Plot the matrix of scalar products as an image
thr = -40; % lower threshold in dB 
figure(6)
imagesc(max(10*log10(abs(SpA)),thr))
colorbar
xlabel('Atom index')
ylabel('Atom index')
title('Scalar products in dB - analysis dictionary')
figure(7)
imagesc(max(10*log10(abs(SpS)),thr))
colorbar
xlabel('Atom index')
ylabel('Atom index')
title('Scalar products in dB - synthesis dictionary')

